(*
  CamlQuizz, an IRC quizbot
  Copyright (C) 2004 Samuel Mimram, Julien Cristau, Brice Goglin

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2, as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.
*)
(* $Id: superquizz.ml,v 1.101 2005/06/06 12:36:25 julio Exp $ *)

open Irc

let very_very_verbose = ref false

let fatal_error err =
  Printf.eprintf "CamlQuizz fatal error: %s\nExiting...\n%!" err;
  exit 1

let () =
  Printf.printf "CamlQuizz %s\n\n" Settings.version;
  Random.self_init();
  let apply_args = ref (fun () -> ()) in
  let set_arg arg v =
    let old_apply_args = !apply_args in
      apply_args := (fun () -> arg := v; old_apply_args ())
  in
  Arg.parse
    [
      "-s", Arg.String (set_arg Settings.server), "       Server name";
      "-p", Arg.Int (set_arg Settings.port), "       Server port";
      "-n", Arg.String (set_arg Settings.nick), "       Nick";
      "-c", Arg.String (set_arg Settings.chan), "       Chan";
      "-q", Arg.String (set_arg Settings.database), "       Question file";
      "--team-scores", Arg.String (set_arg Settings.team_scores), " Team scores";
      "--nick-scores", Arg.String (set_arg Settings.nick_scores),
      " Individual scores";
      "--pass", Arg.String (set_arg Settings.password), "   Nickserv password";
      "--config", Arg.Set_string Settings.config_file, " Configuration file";
      "-vvv", Arg.Unit (fun () -> set_arg very_very_verbose true), "     Be very very verbose";
    ]
    (fun _ -> ())
    ("Usage: " ^ Sys.argv.(0) ^ " [options]");
    Settings.read_conf !Settings.config_file;
    !apply_args ()

module C = Client
  (
    struct
      let server = Settings.get_server ()
      let port = Settings.get_port ()
      let nick = Settings.get_nick ()
      let user = Settings.get_user ()
      let host = Settings.get_host ()
      let realname = Settings.get_realname ()
    end
  )

let get_questions () =
  let c =
    try
      open_in (Settings.questions_file())
    with
      | Sys_error _ -> fatal_error ("Could not find the given question file (" ^ Settings.questions_file() ^ ").")
  in
  let list = ref [] in
    (
      try
	while true do
	  (* each line of the database file consists of
	     "question \ answer"
	  *)
	  let line = input_line c in
	  let split = String.index line '\\' in
	  let q,a = (String.sub line 0 (split - 1), String.sub line (split + 2)
		       (String.length line - split - 3)) in
	    list := (q, a):: !list
	done
      with End_of_file -> close_in c
    );
    Array.of_list !list

let questions = ref (get_questions ())

let quizz_start_time = ref 0.

module Lowercase_hash = struct
  type t = string
  let equal a b = String.lowercase a = String.lowercase b
  let hash a = Hashtbl.hash(String.lowercase a)
end
module Lowercase_hashtbl = Hashtbl.Make(Lowercase_hash)

(* initialize useful structures... *)
let team_scores = Lowercase_hashtbl.create 1000
let nick_scores = Lowercase_hashtbl.create 1000
let teams = Lowercase_hashtbl.create 100
let ops = ref []

(* ...and functions to manipulate them *)
let add_nick_to_team = Lowercase_hashtbl.replace teams
let forget_team_of_nick = Lowercase_hashtbl.remove teams
let has_team = Lowercase_hashtbl.mem teams
let team s = try Lowercase_hashtbl.find teams s with Not_found -> s

let isop s = List.mem s !ops

let nick_score s =
  try Lowercase_hashtbl.find nick_scores s
  with Not_found -> 0, 0

let team_score s =
  try Lowercase_hashtbl.find team_scores s
  with Not_found -> 0, 0

(* nick [s] has won the last question *)
let incr_score s streak team_streak =
  let n,m = nick_score s in
  Lowercase_hashtbl.replace nick_scores s (n+1, max m streak);
  let t = team s in
  let n,m = team_score t in
  Lowercase_hashtbl.replace team_scores t (n+1, max m team_streak)

(* read scores from a file and feed the hashtbl *)
let init_scores () =
  begin
  let buf =
    try
      Scanf.Scanning.from_file (Settings.team_scores_file())
    with
      | Sys_error _ -> fatal_error ("Could not find the given score file (" ^ Settings.team_scores_file() ^ ").")
  in
  let scan () = Scanf.bscanf buf "%s\t%d\t%d\n" (fun s i j -> s, i, j) in
    try
      while true do
	let s,i,j = scan() in
	  Lowercase_hashtbl.replace team_scores s (i, j)
      done
    with End_of_file -> ()
      | e -> Printf.eprintf "Badly formatted team_scores file\n"; raise e
  end;
  let buf =
    try
      Scanf.Scanning.from_file (Settings.nick_scores_file())
    with
      | Sys_error _ -> fatal_error ("Could not find the given score file (" ^ Settings.nick_scores_file() ^ ").")
  in
  let scan () = Scanf.bscanf buf "%s\t%d\t%d\n" (fun s i j -> s, i, j) in
    try
      while true do
	let s,i,j = scan() in
	  Lowercase_hashtbl.replace nick_scores s (i, j)
      done
    with End_of_file -> ()
      | e -> Printf.eprintf "Badly formatted nick_scores file\n"; raise e


(* save all scores to the file *)
let dump_scores file tbl =
  let tempfile,c = Filename.open_temp_file "superquizz" ".tmp" in
    Lowercase_hashtbl.iter
      (fun s (i, j) -> Printf.fprintf c "%s\t%d\t%d\n" s i j) tbl;
    close_out c;
    Sys.rename tempfile file

let get_top n tbl =
  let ans = Array.make n (-1, "") in
  let get_rank s =
    let r = ref n in
      for i = (n - 1) downto 0
      do
	if s > fst ans.(i) then r := i
      done; !r
  in
  let insert team pos =
    if pos < n then
      (
	if pos < n - 1 then
	  for i = n - 2 downto pos
	  do
	    ans.(i + 1) <- ans.(i)
	  done;
	ans.(pos) <- team
      )
  in
    Lowercase_hashtbl.iter (fun s (i, j) -> insert (i, s) (get_rank i)) tbl;
    Array.map snd ans

(* find a random question and return (question,answer) *)
let pick_question () =
  let nb = Array.length(!questions) in
  let n = Random.int nb in
  let q, a = !questions.(n) in
    (
      if q = "#S" then
	let q = String.copy a in
	let swap i j =
	  let tmp = q.[i] in
	    q.[i] <- q.[j];
	    q.[j] <- tmp
	in
	  for i = 0 to ((String.length q) - 1)
	  do
	    swap i (Random.int (String.length q))
	  done;
	  "Mot mlang : " ^ (String.uppercase q)
      else
	q
    ), a

(* remove leading articles, ... from the [truc] string *)
let minimize truc =
  let s = ref truc in
  let rec aux () =
    let ignore = Settings.ignored () in
      for i = 0 to Array.length ignore - 1 do
	if String.length !s >= String.length ignore.(i) &&
	  String.sub !s 0 (String.length ignore.(i)) = ignore.(i) then
	    begin
	      s := String.sub !s (String.length ignore.(i))
		(String.length !s - String.length ignore.(i)); aux ()
	    end
      done
  in aux(); !s

let print_clue a =
  if String.length a = 1 then "" else
    (bicolor_text 15 2 "Une petite aide: ") ^
    (bicolor_text 8 2
       (let l = if String.length a = 2 then 3 else String.length a in
	  if Random.bool() then
	    String.sub a 0 (l / 3) ^ "..."
	  else
	    "..." ^ String.sub a (String.length a - (l / 3)) (l / 3)
       )
    )

(* return [true] if [user_string] is considered as equivalent to the
   right answer *)
let test_answer answer user_string =
  Utils.split_string (minimize (Utils.unaccent(String.lowercase answer))) =
  Utils.split_string (minimize (Utils.unaccent(String.lowercase user_string)))


let print_intro dst =
  C.say dst
    ((bicolor_text 8 2 (bold_text " "))
     ^ (bicolor_text 0 2 (bold_text "Caml Quizz"))
     ^ (bicolor_text 8 2 (bold_text " ")));
  C.say dst
    ((bicolor_text 0 2 (bold_text "Il y a "))
     ^ (bicolor_text 4 2 (bold_text (string_of_int (Array.length !questions))))
     ^ (bicolor_text 0 2 (bold_text " questions dans la base.")));
  C.say dst
    (bold_text (bicolor_text 15 2 "En attente de joueurs, tapez " ^
		  bicolor_text 9 2 "!quizz" ^
		  bicolor_text 15 2 " pour lancer le Quizz!"))

let print_cmds dst =
      C.say dst
	(bicolor_text 15 2
           (bold_text "Liste des commandes disponibles pour le Quizz :"));
      C.say dst
	((bicolor_text 0 2 (bold_text "!repete"))
         ^ (bicolor_text 3 2 (bold_text "  pour rpter la question pose")));
      C.say dst
	((bicolor_text 0 2 (bold_text "!score"))
         ^ (bicolor_text 3 2 (bold_text "  pour connaitre son score ou le score d'un autre")));
      C.say dst
	((bicolor_text 0 2 (bold_text "!top10"))
         ^ (bicolor_text 15 2 (bold_text " ou "))
         ^ (bicolor_text 0 2 (bold_text "!top3"))
         ^ (bicolor_text 3 2 (bold_text "  pour afficher les scores des meilleurs joueurs")));
      C.say dst
	((bicolor_text 0 2 (bold_text "!ajouter"))
         ^ (bicolor_text 3 2 (bold_text "  pour proposer une nouvelle question/rponse")));
      C.say dst
	((bicolor_text 0 2 (bold_text "!nick x"))
         ^ (bicolor_text 3 2 (bold_text "  pour choisir votre quipe de jeu")))


let print_help dst =
      C.say dst
        ((bicolor_text 8 2 (bold_text " "))
         ^ (bicolor_text 15 2 (bold_text "Bienvenue dans "))
         ^ (bicolor_text 0 2 (bold_text ("CamlQuizz " ^ Settings.version)))
         ^ (bicolor_text 15 2 (bold_text " - "))
         ^ (bicolor_text 4 2 (bold_text "by some crazy geeks from the ENS Lyon"))
         ^ (bicolor_text 8 2 (bold_text " ")) );
      C.say dst
        ((bicolor_text 0 2 (bold_text
          "Vous devez taper le premier la bonne rponse avec l'orthographe correcte.")));
      C.say dst
        ((bicolor_text 0 2 (bold_text
          "Les accents, ainsi que les articles et conjonctions en dbut de rponse sont optionnels (ex: Dans la fort = foret)")));
      C.say dst
        ((bicolor_text 0 2 (bold_text
          "Les grands nombres,  l'exception des annes, doivent tre taps avec un espace comme sparateur des milliers (ex: 35 000)")));
      C.say dst
        ((bicolor_text 0 2 (bold_text
          "Plus la rponse est longue, plus il y a de lettres donnes dans l'aide.")));
      C.say dst
        ((bicolor_text 0 2 (bold_text
          "Tapez !commandes pour connaitre les commandes que le Quizz reconnait")))

let print_top10 dst =
  let rank = get_top 10 nick_scores in
  let score = Array.map (fun e -> fst (nick_score e)) rank in
    C.say dst
      (bold_text
	 (bicolor_text 15 2 "Classement des "
	  ^ bicolor_text 4 2 "10 meilleurs joueurs"
	  ^ bicolor_text 15 2 " : "
	  ^ underline_text
	    (bicolor_text 0 2 "01." ^
	     bicolor_text 3 2
	       (" " ^ rank.(0) ^ " (" ^ string_of_int (score.(0)) ^ ")"))));
(* may be factorized *)
    C.say dst
      (bold_text
	 (bicolor_text 15 2 "02."
	  ^ bicolor_text 3 2
            (" " ^ rank.(1) ^ " (" ^ string_of_int (score.(1)) ^ ")")
	  ^ bicolor_text 15 2 " - 03."
	  ^ bicolor_text 3 2
            (" " ^ rank.(2) ^ " (" ^ string_of_int (score.(2)) ^ ")")
	  ^ bicolor_text 15 2 " - 04."
	  ^ bicolor_text 3 2
            (" " ^ rank.(3) ^ " (" ^ string_of_int (score.(3)) ^ ")")));
    C.say dst
      (bold_text
	 (bicolor_text 15 2 "05."
	  ^ bicolor_text 3 2
            (" " ^ rank.(4) ^ " (" ^ string_of_int (score.(4)) ^ ")")
	  ^ bicolor_text 15 2 " - 06."
	  ^ bicolor_text 3 2
            (" " ^ rank.(5) ^ " (" ^ string_of_int (score.(5)) ^ ")")
	  ^ bicolor_text 15 2 " - 07."
	  ^ bicolor_text 3 2
            (" " ^ rank.(6) ^ " (" ^ string_of_int (score.(6)) ^ ")")));
    C.say dst
      (bold_text
	 (bicolor_text 15 2 "08."
	  ^ bicolor_text 3 2
            (" " ^ rank.(7) ^ " (" ^ string_of_int (score.(7)) ^ ")")
	  ^ bicolor_text 15 2 " - 09."
	  ^ bicolor_text 3 2
            (" " ^ rank.(8) ^ " (" ^ string_of_int (score.(8)) ^ ")")
	  ^ bicolor_text 15 2 " - 10."
	  ^ bicolor_text 3 2
            (" " ^ rank.(9) ^ " (" ^ string_of_int (score.(9)) ^ ")")))

let print_top3 dst =
  let rank = get_top 3 team_scores in
  let score = Array.map (fun e -> fst (team_score e)) rank in
    C.say dst
      (bold_text
	 (bicolor_text 15 2 "Les "
	  ^ bicolor_text 4 2 "3 meilleures quipes"
       ^ bicolor_text 15 2 " : "
       ^ underline_text
	    (bicolor_text 0 2 "1."
	     ^ bicolor_text 3 2
	       (" " ^ rank.(0) ^ " (" ^ string_of_int (score.(0)) ^ ")"))
       ^ bicolor_text 15 2 " - 2."
       ^ bicolor_text 3 2
	    (" " ^ rank.(1) ^ " (" ^ string_of_int (score.(1)) ^ ")")
       ^ bicolor_text 15 2 " - 3."
       ^ bicolor_text 3 2
	    (" " ^ rank.(2) ^ " (" ^ string_of_int (score.(2)) ^ ")")))

let active_duration () =
  let d = int_of_float ((Unix.time ()) -. !quizz_start_time) in
  let sec = d mod 60 in
  let d = d / 60 in
  let min = d mod 60 in
  let d = d / 60 in
    Printf.sprintf "%dhrs %dmins %dsecs" d min sec

let question_counter = ref 0
let success_counter = ref 0
let consecutive_unanswered = ref 0

let success_ratio_string () =
  let ratio =
    (float_of_int (!success_counter*100)) /.
      (float_of_int (
       if (!question_counter != 0) then !question_counter else 1
      )) in
    ((bicolor_text 0 2 "[") ^
     (bicolor_text 4 2 (Printf.sprintf "%.1f%%" ratio)) ^
     (bicolor_text 0 2 "]: "))

let print_question q =
  C.say (Settings.get_chan ())
    (bold_text
       ((success_ratio_string ()) ^ (bicolor_text 8 2 q)))


let last_winner = ref ""
let last_winner_count =  ref 0
let last_winner_team = ref ""
let last_winner_team_count =  ref 0
let pub = ref true

exception Start_quizz
exception Good_answer of string

type loop_states =
  | Loop_not_started
  | Quizz_starting15
  | Loop_starting15
  | Loop_starting10
  | Loop_starting5
  | Loop_before_clue of string*string
  | Loop_after_clue of string*string
  | Loop_ending
let current_loop_state = ref Loop_not_started

let signal_handle signum =
    match !current_loop_state with
      | Quizz_starting15 ->
          current_loop_state := Loop_starting5;
          ignore (Unix.alarm (Settings.get_delay_starting5 ()));
          print_top3 (Settings.get_chan())
      | Loop_starting15 ->
          current_loop_state := Loop_starting10;
          ignore (Unix.alarm (Settings.get_delay_starting10 ()
			      - Settings.get_delay_starting5 ()));
          if ( !question_counter mod 50 = 0 ) then
            print_top10 (Settings.get_chan())
          else if ( !question_counter mod 25 = 0 ) then
            print_help (Settings.get_chan())
          else if ( !question_counter mod 5 = 0) then
            print_top3 (Settings.get_chan())
      | Loop_starting10 ->
          current_loop_state := Loop_starting5;
          ignore (Unix.alarm (Settings.get_delay_starting5 ()));
          let m = bicolor_text 4 2
		    ("Le Quizz continue dans "
                     ^ (string_of_int (Settings.get_delay_starting5 ()))
                     ^ " secondes...") in
            if ( !last_winner = "" ) then
              C.say (Settings.get_chan ()) (bold_text m)
            else
              let s = snd(nick_score(!last_winner)) in
                C.say (Settings.get_chan ()) (bold_text
						(m ^ (bicolor_text 15 2 " (Dernier gagnant: ") ^
						 (bicolor_text 3 2 !last_winner) ^
						 (bicolor_text 15 2 " Maximum: ") ^
						 (bicolor_text 3 2 (string_of_int s ^ " d'affile" )) ^
						 (bicolor_text 15 2 ")")))
      | Loop_starting5 ->
          (* poser une question *)
	  let q, a = pick_question () in
	    print_question q;
            current_loop_state := Loop_before_clue(q, a);
            ignore (Unix.alarm (Settings.get_delay_before_clue ()));
            incr question_counter;
      | Loop_before_clue(q,a) ->
          (* afficher un indice *)
          current_loop_state := Loop_after_clue(q,a);
          ignore (Unix.alarm (Settings.get_delay_after_clue ()));
          C.say (Settings.get_chan ()) (bold_text
					  ((bicolor_text 4 2 "Le dlai est bientt coul! ") ^
					   (print_clue a)));
      | Loop_after_clue(q,a) ->
            (* afficher la reponse *)
            C.say (Settings.get_chan ())
	      (bold_text
		 ((bicolor_text 4 2
		     "Dsol, le dlai est coul pour cette question...") ^
		  (bicolor_text 15 2 " La rponse tait: ") ^
		  (bicolor_text 8 2 a)));
            incr consecutive_unanswered;
            if ( !consecutive_unanswered < Settings.get_unanswered_before_stop () ) then
              begin
                current_loop_state := Loop_ending;
                ignore (Unix.alarm (Settings.get_delay_ending ()));
              end
            else
              begin
                C.say (Settings.get_chan ())
		  (bold_text
		     (
		       (bicolor_text 15 2 "Il y a eu ") ^
		       (bicolor_text 4 2
			  (string_of_int
			     (Settings.get_unanswered_before_stop()))) ^
		       (bicolor_text 15 2 " questions non trouves. ") ^
		       (bicolor_text 4 2 "Le Quizz est suspendu... ") ^
		       (bicolor_text 8 2
                          ("(actif depuis " ^
                           (active_duration ()) ^
                           ")"))
		     )
		  );
                current_loop_state := Loop_not_started;
                ignore (Unix.alarm (Settings.get_delay_between_announce_suspended ()));
              end
      | Loop_ending ->
          (* le quizz reprend dans 15s *)
          C.say (Settings.get_chan ())
	    (bold_text
	       (bicolor_text 4 2
		  ("Le Quizz continue dans " ^
		   (string_of_int (Settings.get_delay_starting15 ()))
		   ^ " secondes...")));
          current_loop_state := Loop_starting15;
          ignore (Unix.alarm (Settings.get_delay_starting15 ()
			      - Settings.get_delay_starting10 ()));
      | Loop_not_started ->
          if !pub then print_intro (Settings.get_chan ());
          ignore (Unix.alarm (Settings.get_delay_between_announce_suspended ()))

let on_ctcp_ping_reply sender receiver timestamp =
  let nick = string_of_sender sender in
  let time = Unix.gettimeofday() in
    C.notice nick
      (
	try
	  Printf.sprintf
	    "(Power-Ping) Tu lag trs prcisment de %.3f secondes" (time -. (float_of_string timestamp))
	with
	    Failure("float_of_string") -> Printf.sprintf "(Power-Ping) Timestamp invalide : \"%s\"" timestamp
      )

let on_mode sender mode =
  match mode with
    | Add_channel_mode (chan, (Channel_mode_op user)) ->
	ops := user :: !ops
    | Remove_channel_mode (chan, (Channel_mode_op user)) ->
	ops := List.filter ((<>) user) !ops
    | _ -> ()


let on_nick sender nick =
  let sender = string_of_sender sender in
  if List.mem sender !ops then
    ops := nick :: List.filter ((<>) sender) !ops;
  if has_team sender then
    let group = team sender in
      begin
	forget_team_of_nick sender;
	add_nick_to_team nick group
      end


let on_reply sender kind params =
  match kind with
    | Rpl_name_reply ->
	let l = Str.split (Str.regexp " ") params.(3) in
	  List.iter
	    (fun s -> if s.[0] = '@' then
	       ops := String.sub s 1 (String.length s - 1) :: !ops) l;
    | _ -> ()

let remove nick =
    ops := List.filter ((<>) nick) !ops;
    forget_team_of_nick nick
let on_quit sender reason =
  remove (string_of_sender sender)
let on_part sender chan reason =
  if String.lowercase chan = String.lowercase(Settings.get_chan()) then
    remove (string_of_sender sender)
let on_kick sender chan user reason =
  if String.lowercase chan = String.lowercase(Settings.get_chan()) then
    remove user

let on_message sender receiver msg =
  if ( msg = "!quizz" ) then
    raise Start_quizz
  else if msg = "!score" then
    let peer = string_of_sender sender in
    let i,j = nick_score peer in
    let k,l = team_score (team peer) in
      C.notice peer
	(Printf.sprintf "Votre score: %d rponses correctes. Maximum d'affile: %d rponses. Score de votre quipe: %d rponses dont %d d'affile" i j k l)
  else if String.length msg > 6 && (String.sub msg 0 6 = "!score") then
    let remainder = String.sub msg 6 (String.length msg - 6) in
      if remainder.[0] != ' ' then () else
	let list = Str.split (Str.regexp " +") remainder in
	let peer =
	  if list = [] then string_of_sender sender
	  else List.hd list
	in
	let i,j = nick_score peer in
	  C.notice (string_of_sender sender)
	    (Printf.sprintf "Score de %s: %d rponses correctes. Maximum d'affile: %d rponses" peer i j)
  else if (msg = "!top3") then
    print_top3 (string_of_sender sender)
  else if (msg = "!top10") then
    print_top10 (string_of_sender sender)
  else if (msg = "!aide") then
    print_help (string_of_sender sender)
  else if (msg = "!commandes") then
    print_cmds (string_of_sender sender)
  else if String.length msg >= 7 && String.sub msg 0 6 = "!nick " then
    let remainder = String.sub msg 6 (String.length msg - 6) in
    let list = Str.split (Str.regexp " +") remainder in
      if list = [] then ()
      else begin
	let nick = string_of_sender sender in
	let group = List.hd list in
	let i,j = team_score group in
	  add_nick_to_team nick group;
	  C.notice nick
	    ("Votre nick de jeu est maintenant " ^ group ^ ". " ^
	     "Votre score: " ^ string_of_int i ^ " rponses correctes. " ^
	     "Maximum d'affile: " ^ string_of_int j ^ " rponses")
      end
  else if (msg = "!repete") then
    match !current_loop_state with
      | Loop_not_started
      | Loop_ending
      | Quizz_starting15
      | Loop_starting15
      | Loop_starting10
      | Loop_starting5 -> ()
      | Loop_before_clue(q,a)
      | Loop_after_clue(q,a) ->
	  print_question q
  else if (msg = "!ping") then
    let timestamp = Unix.gettimeofday() in
    let nick = string_of_sender sender in
      C.ctcp_ping nick timestamp
  else if msg = "!stop" then begin
    match !current_loop_state with
      | Loop_not_started ->
	  if !pub then begin
	    C.notice (string_of_sender sender) "OK, j'arrte la pub";
	    pub := false
	  end
      | _ ->
	  if isop (string_of_sender sender) then begin
	    current_loop_state := Loop_not_started;
	    C.say (Settings.get_chan())
	      (bold_text
		 (bicolor_text 4 2
		    "Le Quizz est termin pour le moment... " ^
		  bicolor_text 8 2
		    ("(actif depuis " ^ active_duration() ^ ")" )
		 )
	      )
	  end
  end;
  begin if String.lowercase receiver = String.lowercase(Settings.get_chan())
  then
    match !current_loop_state with
      | Loop_not_started
      | Loop_ending
      | Quizz_starting15
      | Loop_starting15
      | Loop_starting10
      | Loop_starting5 -> ()
      | Loop_before_clue(q,a)
      | Loop_after_clue(q,a) ->
	  if test_answer a msg then
	    raise (Good_answer (string_of_sender sender))
  end

let () =
  C.dump_everything := !very_very_verbose;
  C.on_message := on_message;
  C.on_mode := on_mode;
  C.on_nick := on_nick;
  C.on_reply := on_reply;
  C.on_ctcp_ping_reply := on_ctcp_ping_reply;
  C.on_part := on_part;
  C.on_kick := on_kick;
  C.on_quit := on_quit;
  init_scores();
  while true do
    C.connect ();
    begin match Settings.get_password() with
      | Some password -> C.say "NickServ" ("identify " ^ password)
      | None -> ()
    end;
    C.join (Settings.get_chan ());
    Sys.set_signal (Sys.sigalrm) (Sys.Signal_handle (signal_handle));
    ignore (Unix.alarm (Settings.get_delay_between_announce_suspended ()));
    print_intro (Settings.get_chan ());
    try
      while true do
	try
          C.event_loop ()
	with
          | Start_quizz ->
	      (
		match !current_loop_state with
                  | Loop_not_started ->
                      C.say (Settings.get_chan ())
			(bold_text (bicolor_text 4 2
				      ("Le Quizz dmarre dans " ^
				       (string_of_int (Settings.get_delay_starting15 ()))
				       ^ " secondes, prparez-vous!" )));
		      quizz_start_time := Unix.time ();
		      question_counter := 0;
		      success_counter := 0;
		      consecutive_unanswered := 0;
                      current_loop_state := Quizz_starting15;
		      pub := true;
                      ignore (Unix.alarm (Settings.get_delay_starting15 ()
                                          - Settings.get_delay_starting5 ())); (* but keep the full delay *)
                  | _ ->
                      (* C.say (Settings.get_chan ())
			 ("Un quizz est dj commenc..."); *) ()
	      )
          | Good_answer nick ->
	      (
		consecutive_unanswered := 0;
		match !current_loop_state with
                  | Loop_not_started
                  | Loop_ending
                  | Quizz_starting15
                  | Loop_starting15
                  | Loop_starting10
                  | Loop_starting5 -> ()
                  | Loop_before_clue(q, a)
                  | Loop_after_clue(q, a) ->
                      let winner = team nick in
			if String.lowercase !last_winner = nick then
			  incr last_winner_count
			else begin
			  last_winner := nick;
			  last_winner_count := 1;
			end;
			if String.lowercase !last_winner_team =
			  String.lowercase winner then
                            incr last_winner_team_count
			else begin
                          last_winner_team := winner;
                          last_winner_team_count := 1;
			end;
			incr_score nick !last_winner_count
			  !last_winner_team_count;
			let s = fst (team_score winner) in
			let m = bicolor_text 9 2 "Correct!" ^
				bicolor_text 15 2 " La rponse est: " ^
				bicolor_text 8 2 a ^
				bicolor_text 15 2 ". Continue comme a " ^
				bicolor_text 9 2 winner ^
				bicolor_text 15 2 " ! (" ^
				bicolor_text 9 2 (string_of_int s) ^
				bicolor_text 15 2 " rponses)"
			in
			  if (!last_winner_team_count > 1) then
			    C.say (Settings.get_chan ())
 			      (bold_text
				 (m ^ bicolor_text 15 2 " dont "
				  ^ bicolor_text 9 2
				    (string_of_int !last_winner_team_count)
				  ^ bicolor_text 15 2 " d'affile!"))
			  else
			    C.say (Settings.get_chan ()) (bold_text m);
			  C.change_mode (Add_channel_mode ((Settings.get_chan ()), Channel_mode_voice nick));
			  dump_scores (Settings.nick_scores_file())
			    nick_scores;
			  dump_scores (Settings.team_scores_file())
			    team_scores;
			  incr success_counter;
			  current_loop_state := Loop_ending;
			  ignore (Unix.alarm (Settings.get_delay_ending ()));
	      )
      done
    with
      | End_of_file ->
	  C.quit "Je reviens tout de suite!";
	  C.disconnect ()
  done
